import QtQuick 2.9
import QtQuick.Controls 2.2
import MuseScore 3.0
// fork of the MIT licensed plugin: Parking B - Tap Tempo https://musescore.org/node/331133
MuseScore {
  menuPath: "Plugins.Tempo "+pn[pi].pname
  version: "20231125A"
  property var pi:0
  property var pn:[
     {pname:'Speed Up'                        ,shiftnumber:'+1'   },
     {pname:'Slow Down'                       ,shiftnumber:'-1'  },
     {pname:'Double note value'               ,switchnote:'-2'   },
     {pname:'Increase note value by dot'      ,switchnote:'-1'   },
     {pname:'Decrease note value by dot'      ,switchnote:'+1'   },
     {pname:'Half note value'                 ,switchnote:'+2'     },
  ]
  id: pluginscope
  Component.onCompleted:{
    if (mscoreMajorVersion >= 4) {
      // pluginscope.pluginType="dialog"
      pluginscope.title = 'QT_'+pn[pi].pname
      // some_id.thumbnailName = "thumbnail.png";
      // some_id.categoryCode = "some_caregory";
    }
  }
  property var multipliers:[
    //mult is a tempo-multiplier compared to a crotchet      
    {text: '\uECA2',               mult: 4     , sym: '<sym>metNoteWhole</sym>'                                                                         }, // 1/1
    {text: '\uECA3 \uECB7',        mult: 3     , sym: '<sym>metNoteHalfUp</sym><sym>metAugmentationDot</sym>'                                           }, // 1/2.
    {text: '\uECA3',               mult: 2     , sym: '<sym>metNoteHalfUp</sym>'                                                                        }, // 1/2
    {text: '\uECA5 \uECB7 \uECB7', mult: 1.75  , sym: '<sym>metNoteQuarterUp</sym><sym>metAugmentationDot</sym><sym>metAugmentationDot</sym>' ,dbldot:1 }, // 1/4..
    {text: '\uECA5 \uECB7',        mult: 1.5   , sym: '<sym>metNoteQuarterUp</sym><sym>metAugmentationDot</sym>'                                        }, // 1/4.
    {text: '\uECA5',               mult: 1     , sym: '<sym>metNoteQuarterUp</sym>'                                                                     }, // 1/4
    {text: '\uECA7 \uECB7 \uECB7', mult: 0.875 , sym: '<sym>metNote8thUp</sym><sym>metAugmentationDot</sym><sym>metAugmentationDot</sym>'     ,dbldot:1 }, // 1/8..
    {text: '\uECA7 \uECB7',        mult: 0.75  , sym: '<sym>metNote8thUp</sym><sym>metAugmentationDot</sym>'                                            }, // 1/8.
    {text: '\uECA7',               mult: 0.5   , sym: '<sym>metNote8thUp</sym>'                                                                         }, // 1/8
    {text: '\uECA9 \uECB7 \uECB7', mult: 0.4375, sym: '<sym>metNote16thUp</sym><sym>metAugmentationDot</sym><sym>metAugmentationDot</sym>'    ,dbldot:1 }, //1/16..
    {text: '\uECA9 \uECB7',        mult: 0.375 , sym: '<sym>metNote16thUp</sym><sym>metAugmentationDot</sym>'                                           }, //1/16.
    {text: '\uECA9',               mult: 0.25  , sym: '<sym>metNote16thUp</sym>'                                                                        }, //1/16
  ]
  property var switchnote:function(m,d){
    var nodbldots=multipliers.filter(function(x){return !x.dbldot}) /*not support double dotted tempo marking yet*/
    var ni=nodbldots.findIndex(function(e){return e.mult===m})
    var n=~ni&&nodbldots[ni+d]
    return n?n.mult:m
  }
  property var shiftnumber:function(t,d){
    var metros='40,42,44,46,48,50,52,54,56,58,60,63,66,69,72,76,80,84,88,92,96,100,104,108,112,116,120,126,132,138,144,152,160,168,176,184,192,200,208'.split(',').map(function(s){return Number(s)})
    var metrolevel=function(t){return metros.reduce(function (a,b){return b<=t?b:a })}
    var metroindex=function(t){return metros.indexOf(metrolevel(t))}
    var validIndex=function(i){
      return   i<0                    ? 0
             :       metros.length<=i ? metros.length-1
             :                          i
    }
    return metros[validIndex(metroindex(t)+d)]
  }
  onRun:{
    var map=function(o,f){var x=[];for(var i in o){x.push(f(o[i],Number(i)))};return x}
    map(curScore.selection.elements,function(e,i){
      var tempoElement
      var curSegment
      var skip
      if (e){
        if (e.type == Element.TEMPO_TEXT) {
          tempoElement = e
        } else if(i===0){
          // chord/note
          curSegment = e.parent
          while (curSegment.type !== Element.SEGMENT) {
            curSegment = curSegment.parent
          }
          tempoElement = findExistingTempoElement(curSegment)
        } else {skip=true}
      }
      if(!skip){
        var tempomult=1
        var tempo=120
        if(!tempoElement){
        }else{
          tempomult=findBeatBaseFromMarking(tempoElement)
          if(tempoElement.tempoFollowText==false || !~tempomult){
            tempomult=1
            tempo=tempoElement.tempo*60/*ms3 default tempo:2 = 120 crotchet BPM*/
          }else{
            tempo=tempoElement?Number(tempoElement.text.match(/=.*?(\d+)/)[1]):120
          }
          if(pn[pi].shiftnumber)
            tempo=shiftnumber(tempo,+pn[pi].shiftnumber)
          if(pn[pi].switchnote){
            var t0=tempomult
            tempomult=switchnote(tempomult,+pn[pi].switchnote)
            tempo*=t0/tempomult
          }
        }
        var tempotext=multipliers.find(function(e) { return e.mult===tempomult}).sym+' = '+tempo.toFixed(0)
        curScore.startCmd()
        if (tempoElement) {
          tempoElement.text=tempotext
        } else {
          tempoElement=newElement(Element.TEMPO_TEXT)
          tempoElement.text=tempotext
          var cursor = curScore.newCursor()
          cursor.rewindToTick(curSegment.tick)
          cursor.add(tempoElement)
        }
        if (tempoElement) {
          //changing of tempo can only happen after being added to the segment
          tempoElement.tempo = tempo * tempomult
          tempoElement.tempoFollowText = true //allows for manual fiddling by the user afterwards
        }
        curScore.endCmd()
      }
    })
  }
  function findExistingTempoElement(segment) { //look in reverse order, there might be multiple TEMPO_TEXTs attached
    // in that case MuseScore uses the last one in the list
    for (var i = segment.annotations.length; i-- > 0; ) {
      if (segment.annotations[i].type === Element.TEMPO_TEXT) {
        return (segment.annotations[i]);
      }
    }
    return undefined; //invalid - no tempo text found
  }
  function findBeatBaseFromMarking(tempoMarking) {    //fork return
    // First look for metronome marking symbols
    var foundTempoText=tempoMarking.text.replace('<sym>space</sym>', '');
    var foundMetronomeSymbols = foundTempoText.match(/(<sym>met.*<\/sym>)+/g);

    // strip html tags and split around '='
    var data = foundTempoText.replace(/<.*?>/g,'').split('=');
    var tempo=parseInt(data[1]);
    if (isNaN(tempo)) tempo=0;


    if (foundMetronomeSymbols !== null) {
      // Locate the index in our dropdown matching the found beatString
      for (var i = multipliers.length; --i >= 0; ) {
          if (multipliers[i].sym == foundMetronomeSymbols[0]) {
              // Found this marking in the dropdown at metronomeMarkIndex
              return multipliers[i].mult
          }
      }
    } else {
      // Metronome marking symbols are substituted with their character entity if the text was edited
      // UTF-16 range [\uECA0 - \uECB6] (double whole - 1024th)
      for (var beatString, charidx = 0; charidx < foundTempoText.length; charidx++) {
          beatString = foundTempoText[charidx];
          if ((beatString >= "\uECA2") && (beatString <= "\uECA9")) {
              // Found base tempo - continue looking for augmentation dots
              while (++charidx < foundTempoText.length) {
                  if (foundTempoText[charidx] == "\uECB7") {
                      beatString += " \uECB7";
                  } else if (foundTempoText[charidx] != ' ') {
                      break; // No longer augmentation dots or spaces
                  }
              }
              // Locate the index in our dropdown matching the found beatString

              for (var i = multipliers.length; --i >= 0; ) {
                  if (multipliers[i].text == beatString) {
                      // Found this marking in the dropdown at metronomeMarkIndex
                      return multipliers[i].mult
                  }
              }

              break; // Done processing base tempo
          }
      }
    }
    return -1
  }
}